#'---
#'title: "Compilation Report | Corona-Rechtsprechung des Bundesverfassungsgerichts"
#'author: Seán Fobbe
#'papersize: a4
#'geometry: margin=3cm
#'fontsize: 11pt
#'output:
#'  pdf_document:
#'    toc: true
#'    toc_depth: 3
#'    number_sections: true
#'    pandoc_args: --listings
#'    includes:
#'      in_header: General_Source_TEX_Preamble_DE.tex
#'      before_body: [BVerfG-Corona_Source_TEX_CompilationTitle.tex]
#'bibliography: packages.bib
#'nocite: '@*'
#' ---




#'\newpage
#+
#'# Einleitung
#'
#+
#'## Überblick
#' Dieses R-Skript lädt den Corpus der Entscheidungen des Bundesverfassungsgerichts (CE-BVerfG) herunter, untersucht ihn auf mit SARS-CoV-2 assoziiertem Vokabular und speichert relevante Entscheidungen.  Es ist die Grundlage für den Datensatz \textbf{\datatitle\ (\datashort )}.
#'
#' Alle mit diesem Skript erstellten Datensätze werden dauerhaft kostenlos und urheberrechtsfrei auf Zenodo, dem wissenschaftlichen Archiv des CERN, veröffentlicht. Alle Versionen sind mit einem persistenten Digital Object Identifier (DOI) versehen. Die neueste Version des Datensatzes ist immer über den Link der Concept DOI erreichbar: \dataconcepturldoi



#+
#'## Funktionsweise

#' Primäre Endprodukte des Skripts sind folgende ZIP-Archive:
#' \begin{enumerate}
#' \item Der Corona-Datensatz im PDF-Format
#' \item Der Corona-Datensatz im TXT-Format
#' \item Alle Analyse-Ergebnisse (Tabellen als CSV, Grafiken als PDF und PNG)
#' \item Der Source Code und alle weiteren Quelldaten
#' \end{enumerate}
#'
#' Zusätzlich werden für alle ZIP-Archive kryptographische Signaturen (SHA2-256 und SHA3-512) berechnet und in einer CSV-Datei hinterlegt. Es kann optional ein PDF-Bericht erstellt werden (siehe unter "Kompilierung").


#+
#'## Systemanforderungen
#' Das Skript in seiner veröffentlichten Form kann nur unter Linux ausgeführt werden, da es Linux-spezifische Optimierungen (z.B. Fork Cluster) und Shell-Kommandos (z.B. OpenSSL) nutzt. Das Skript wurde unter Fedora Linux entwickelt und getestet. Die zur Kompilierung benutzte Version entnehmen Sie bitte dem **sessionInfo()**-Ausdruck am Ende dieses Berichts.
#'
#' In der Standard-Einstellung wird das Skript vollautomatisch die maximale Anzahl an Rechenkernen/Threads auf dem System zu nutzen. Wenn die Anzahl Threads (Variable "fullCores") auf 1 gesetzt wird, ist die Parallelisierung deaktiviert.
#'
#' Auf der Festplatte sollten 2 GB Speicherplatz vorhanden sein.
#' 
#' Um die PDF-Berichte kompilieren zu können benötigen Sie das R package **rmarkdown**, eine vollständige Installation von \LaTeX\ und alle in der Präambel-TEX-Datei angegebenen \LaTeX\ Packages.



#'\newpage
#+
#'## Kompilierung

#' Mit der Funktion **render()** von **rmarkdown** kann der **vollständige Datensatz**  kompiliert und das Skript mitsamt seinen Rechenergebnisse in ein gut lesbares PDF-Format überführt werden.
#'
#' Alle Kommentare sind im roxygen2-Stil gehalten. Das Skript kann daher auch **ohne render()** regulär als R-Skript ausgeführt werden. Es wird in diesem Fall kein PDF-Bericht erstellt und Diagramme werden nicht abgespeichert.

#' 
#' Um den vollständigen Datensatz zu kompilieren und einen PDF-Bericht zu erstellen, kopieren Sie bitte alle im Source-Archiv bereitgestellten Dateien in einen leeren Ordner und führen mit R diesen Befehl aus:

#+ eval = FALSE

rmarkdown::render(input = "BVerfG-Corona_Source_CorpusCreation.R",
                  output_file = paste0("BVerfG-Corona_2021-01-08_CompilationReport.pdf"),
                  envir = new.env())





#'# Parameter

#+
#'## Name des Datensatzes
datasetname <- "BVerfG-Corona"


#'## DOI der konkreten Datensatz-Version

doi.version <- "10.5281/zenodo.4459406"


#'## Verzeichnis für Analyse-Ergebnisse
#' Hinweis: Muss mit einem Schrägstrich enden!
outputdir <- paste0(getwd(), "/ANALYSE/") 



#'## Optionen: Quanteda
tokens_locale <- "de_DE"


#'## Optionen: Knitr

#+
#'### Ausgabe-Format
dev <- c("pdf", "png")

#'### DPI für Raster-Grafiken
dpi <- 300

#'### Ausrichtung von Grafiken im Compilation Report
fig.align <- "center"



#'## Frequenztabellen: Ignorierte Variablen

#' Diese Variablen werden bei der Erstellung der Frequenztabellen nicht berücksichtigt.

varremove <- c("text",
               "eingangsnummer",
               "datum",
               "doc_id",
               "seite",
               "name",
               "ecli",
               "aktenzeichen",
               "tokens",
               "typen",
               "saetze",
               "version")



#'# Vorbereitung

#'## Datumsstempel
#' Dieser Datumsstempel wird in alle Dateinamen eingefügt. Er richtet sich nach der Version des Stamm-Datensatzes.

datestamp <- "2021-01-08"
print(datestamp)


#'## Datum und Uhrzeit (Beginn)

begin.script <- Sys.time()
print(begin.script)


#'## Ordner für Analyse-Ergebnisse erstellen
dir.create(outputdir)


#'## Packages


library(doParallel)   # Parallelisierung
library(ggplot2)      # Fortgeschrittene Datenvisualisierung
library(rmarkdown)    # Wissenschaftliches Reporting
library(knitr)        # Wissenschaftliches Reporting
library(kableExtra)   # Verbesserte Kable Tabellen
library(data.table)   # Fortgeschrittene Datenverarbeitung
library(quanteda)     # Fortgeschrittenes Natural Language Processing



#'## Zusätzliche Funktionen einlesen
#' **Hinweis:** Die hieraus verwendeten Funktionen werden jeweils vor der ersten Benutzung in vollem Umfang angezeigt um den Lesefluss zu verbessern.

source("General_Source_Functions.R")


#'## Quanteda-Optionen setzen
quanteda_options(tokens_locale = tokens_locale)


#'## Knitr Optionen setzen
knitr::opts_chunk$set(fig.path = outputdir,
                      dev = dev,
                      dpi = dpi,
                      fig.align = fig.align)



#'## Vollzitate statistischer Software
knitr::write_bib(c(.packages()), "packages.bib")



#'## Parallelisierung aktivieren
#' Parallelisierung wird zur Datenanalyse mittels **quanteda** und **data.table** verwendet. Die Anzahl Threads wird automatisch auf das verfügbare Maximum des Systems gesetzt, kann aber auch nach Belieben auf das eigene System angepasst werden. Die Parallelisierung kann deaktiviert werden, indem die Variable **fullCores** auf 1 gesetzt wird.
#'
#' Die hier verwendete Funktion **makeForkCluster()** ist viel schneller als die Alternativen, funktioniert aber nur auf Unix-basierten Systemen (Linux, MacOS).

#+
#'### Logische Kerne

fullCores <- detectCores()
print(fullCores)

#'### Quanteda
quanteda_options(threads = fullCores) 

#+
#'### Data.table
setDTthreads(threads = fullCores)  

#'### DoParallel
cl <- makeForkCluster(fullCores)
registerDoParallel(cl)






#'# Stamm-Datensatz einlesen (CE-BVerfG)
#' Der Stamm-Datensatz ist der \enquote{Corpus der Entscheidungen des Bundesverfassungsgerichts} (CE-BVerfG). Dieser enthält alle vom Bundesverfassungsgericht seit 1998 veröffentlichten Entscheidungen. Dessen **aktuellste** Version ist immer über diesen Digital Object Identifier abrufbar: \url{https://doi.org/10.5281/zenodo.3902658}

#+
#'## Download der CSV-Datei
#' Der Datensatz im CSV-Format wird automatisch über einen verschlüsselten und langzeit-stabilen Link aus dem wissenschaftlichen Archiv des CERN heruntergeladen. Dieses Vorgehen garantiert die Verwendung einer authentischen Version des Datensatzes.

download.file("https://zenodo.org/record/4308215/files/CE-BVerfG_2021-01-08_DE_CSV_Datensatz.zip?download=1",
              "Stamm-Datensatz.zip")



#'## ZIP-Archiv entpacken
unzip("Stamm-Datensatz.zip")


#'## ZIP-Archiv löschen
unlink("Stamm-Datensatz.zip")


#'## CSV-Datei einlesen
dt.bverfg <- fread("CE-BVerfG_2021-01-08_DE_CSV_Datensatz.csv")


#'## Korpus-Objekt erstellen
corpus.bverfg <- corpus(dt.bverfg)





#'# Keywords in Context (KWIC)
#' Bei einer KWIC-Analyse (keywords in context) suchen wir nach einem bestimmten Muster und lassen uns die angrenzenden Wörter anzeigen. Konkret wird an dieser Stelle eine alternative Suche nach den Mustern "Corona", "COVID" oder "SARS-CoV" durchgeführt. Groß- und Kleinschreibung wird ignoriert um eventuelle Tippfehler zu erfassen. Das Sichtfenster wird auf 15 Tokens vor und nach dem Treffer gesetzt.


#+
#'## KWIC-Analyse durchführen
kwic <- kwic(corpus.bverfg,
             pattern = "(Corona)|(COVID)|(SARS-CoV)",
             window = 15,
             valuetype = "regex",
             case_insensitive = TRUE)




#'## KWIC-Tabelle speichern

fwrite(data.frame(kwic),
       paste(datasetname,
             datestamp,
             "00_KeywordsInContext.csv",
             sep = "_"))


fwrite(data.frame(kwic),
       paste0(outputdir,
              datasetname,
              "_01_KeywordsInContext.csv"))
       



#'# TXT-Datensatz erstellen

#+
#'## Namen der Corona-Entscheidungen definieren
keep.txt <- unique(kwic$docname)


#'## TXT-Datensatz herunterladen
download.file("https://zenodo.org/record/4308215/files/CE-BVerfG_2021-01-08_DE_TXT_Datensatz.zip?download=1",
              "TXT.zip")

#'## ZIP-Archiv entpacken
unzip("TXT.zip")


#'## Corona-Entscheidungen verpacken
zip(paste(datasetname,
          datestamp,
          "DE_TXT_Datensatz.zip",
          sep = "_"),
    keep.txt)


#'## TXT-Dateien löschen
unlink(list.files(pattern = ".txt"))


#'## ZIP-Archiv löschen
unlink("TXT.zip")




#'# PDF-Datensatz erstellen

#+
#'## Namen der Corona-Entscheidungen definieren
keep.pdf <- gsub(".txt",
                 ".pdf",
                 keep.txt)


#'## PDF-Datensatz herunterladen
download.file("https://zenodo.org/record/4308215/files/CE-BVerfG_2021-01-08_DE_PDF_Datensatz.zip?download=1",
              "PDF.zip")


#'## ZIP-Archiv entpacken
unzip("PDF.zip")


#'## Corona-Entscheidungen verpacken
zip(paste(datasetname,
          datestamp,
          "DE_PDF_Datensatz.zip",
          sep = "_"),
    keep.pdf)


#'## PDF-Dateien löschen
unlink(list.files(pattern = ".pdf"))


#'## ZIP-Archiv löschen
unlink("PDF.zip")




#'# Lexical Dispersion Plot
#'Lexical Dispersion Plots zeigen mit einem vertikalen Strich an, an welcher Stelle in einem Dokument sich ein Token befindet. Alle Dokumente sind auf eine Länge von 1.0 normalisiert, d.h. ein Wert von 0.5 heißt immer, dass sich das Token in der Mitte des jeweiligen Dokumentes befindet. Viele und/oder dicke Striche deuten auf eine große Häufigkeit des Tokens hin.


#+
#'## Quadratisches Format

#+ BVerfG-Corona_02_LexicalDispersion_Quadratisch, fig.height = 12, fig.width = 12
textplot_xray(kwic, scale = "relative")+
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Lexical Dispersion Plot"),
        caption = paste("DOI:",
                        doi.version))+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'## A4-Format

#+ BVerfG-Corona_02_LexicalDispersion_A4, fig.height = 11.7, fig.width = 8.3
textplot_xray(kwic, scale = "relative")+
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Lexical Dispersion Plot"),
        caption = paste("DOI:",
                        doi.version))+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )






#'# Frequenztabellen erstellen



#+
#'## CE-BVerfG auf Corona-Entscheidungen reduzieren

dt.corona <- dt.bverfg[doc_id %in% keep.txt]


#'## Funktion anzeigen

#+ results = "asis"
print(f.fast.freqtable)


#'## Ignorierte Variablen
print(varremove)



#'## Liste zu prüfender Variablen

varlist <- names(dt.corona)

varlist <- setdiff(varlist,
                   varremove)

print(varlist)



#'## Frequenztabellen erstellen

prefix <- paste0(datasetname,
                 "_00_Frequenztabelle_var-")




#+ results = "asis"
f.fast.freqtable(dt.corona,
                 varlist = varlist,
                 sumrow = TRUE,
                 output.list = FALSE,
                 output.kable = TRUE,
                 output.csv = TRUE,
                 outputdir = outputdir,
                 prefix = prefix)



#'# Erstellen der ZIP-Archive

#+
#'## Verpacken der Analyse-Dateien

zip(paste0(datasetname,
           "_",
           datestamp,
           "_DE_",
           basename(outputdir),
           ".zip"),
    basename(outputdir))





#'## Verpacken der Source-Dateien

files.source <- c(list.files(pattern = "Source"),
                  "buttons")


files.source <- grep("spin",
                     files.source,
                     value = TRUE,
                     ignore.case = TRUE,
                     invert = TRUE)

zip(paste(datasetname,
           datestamp,
           "Source_Files.zip",
           sep = "_"),
    files.source)





#'# Kryptographische Hashes
#' Dieses Modul berechnet für jedes ZIP-Archiv zwei Arten von Hashes: SHA2-256 und SHA3-512. Mit diesen kann die Authentizität der Dateien geprüft werden und es wird dokumentiert, dass sie aus diesem Source Code hervorgegangen sind. Die SHA-2 und SHA-3 Algorithmen gelten derzeit als sicher und ein SHA3-Hash mit 512 bit Länge ist nach derzeitigem Wissen auch gegenüber quantenkryptoanalytischen Verfahren hinreichend resistent.

#+
#'## Liste der ZIP-Archive erstellen
files.zip <- list.files(pattern= "\\.zip$",
                        ignore.case = TRUE)


#'## Funktion anzeigen
#+ results = "asis"
print(f.dopar.multihashes)

#'## Hashes berechnen
multihashes <- f.dopar.multihashes(files.zip)


#'## In Data Table umwandeln
setDT(multihashes)



#'## Index hinzufügen
multihashes$index <- seq_len(multihashes[,.N])


#'## In Datei schreiben
fwrite(multihashes,
       paste(datasetname,
             datestamp,
             "KryptographischeHashes.csv",
             sep = "_"),
       na = "NA")


#'## Leerzeichen hinzufügen um Zeilenumbruch zu ermöglichen
multihashes$sha3.512 <- paste(substr(multihashes$sha3.512, 1, 64),
                              substr(multihashes$sha3.512, 65, 128))


#'\newpage
#'## In Bericht anzeigen

kable(multihashes[,.(index,filename)],
      format = "latex",
      align = c("p{1cm}",
                "p{13cm}"),
      booktabs=TRUE,
      longtable=TRUE)


kable(multihashes[,.(index,sha2.256)],
      format = "latex",
      align = c("c",
                "p{13cm}"),
      booktabs=TRUE,
      longtable=TRUE)


#'\newpage

kable(multihashes[,.(index,sha3.512)],
      format = "latex",
      align = c("c",
                "p{13cm}"),
      booktabs=TRUE,
      longtable=TRUE)







#'# Abschluss

#'## Cluster stoppen
stopCluster(cl)


#'## Datum und Uhrzeit (Ende)
end.script <- Sys.time()
print(end.script)

#'## Laufzeit des gesamten Skriptes
print(end.script - begin.script)


#'## Warnungen
warnings()



#'# Parameter für strenge Replikationen

system2("openssl", "version", stdout = TRUE)

sessionInfo()


#'# Literaturverzeichnis

