
#'# Introduction


#+
#'## Overview
#' This proto-package combines a number of functions that I have found useful during the creation of new data sets, particularly text corpora. All of the functions in this package are reasonably well-tested in real-world scenarios.
#'
#' The collection of functions is still in early and very active development. Function names, functionality and requirements may change drastically for some time yet.

#'# f.fast.freqtable: Fast Frequency Tables
#' This function create frequency tables for an arbitrary number of variables. It can return them as a list, write them to an arbitrary folder on disk as CSV files (with an optional prefix and return kable tables that are designed to work well with render() and LaTeX. It is based on data.table and is therefore capable of processing massive data ssets. To show the kable output in render() you must add the Chunk Option "results = 'asis'" when calling the function.
#'

#+
#'## Usage
#' Simply download the R script into the directory of your choice and call it from within R like so:

### source("Package_Source_CorpusCreationTools.R")


#+
#'## Required Arguments


#' @param x A data.table.


#+
#'## Optional arguments


#' @param varlist Character. An optional character vector of variable names to construct tables for. Defaults to all variables.
#' @param sumrow Logical. Whether to add a summary row.
#' @param output.list Logical. Whether to output the frequency tables as a list. Defaults to TRUE. Returns NULL otherwise.
#' @param output.kable Logical. Whether to return kable tables. Defaults to FALSE.
#' @param output.csv Logical. Whether to write CSV files to disk. Defaults to FALSE.
#' @param outputdir Character. The target directory for writing CSV files. Defaults to the current R working directory.
#' @param prefix A string to be added to each CSV file. Default is not to add a string and just to output the variable name as the name of the CSV file. 



#'## Required Packages

library(data.table)
library(knitr)
library(kableExtra)


#'## Function

f.fast.freqtable <- function(x,
                             varlist = names(x),
                             sumrow = TRUE,
                             output.list = TRUE,
                             output.kable = FALSE,
                             output.csv = FALSE,
                             outputdir = "./",
                             prefix = ""){
    
    ## Begin List
    freqtable.list <- vector("list", length(varlist))

    ## Calculate Frequency Table
    for (i in seq_along(varlist)){
        
        varname <- varlist[i]
        
        freqtable <- x[, .N, keyby=c(paste0(varname))]
        
        freqtable[, c("exactpercent",
                      "roundedpercent",
                      "cumulpercent") := {
                          exactpercent  <-  N/sum(N)*100
                          roundedpercent <- round(exactpercent, 2)
                          cumulpercent <- round(cumsum(exactpercent), 2)
                          list(exactpercent,
                               roundedpercent,
                               cumulpercent)}]

        ## Calculate Summary Row
        if (sumrow == TRUE){
            colsums <-  cbind("Total",
                              freqtable[, lapply(.SD, function(x){round(sum(x))}),
                                        .SDcols = c("N",
                                                    "exactpercent",
                                                    "roundedpercent")
                                        ], round(max(freqtable$cumulpercent)))
            
            colnames(colsums)[c(1,5)] <- c(varname, "cumulpercent")
            freqtable <- rbind(freqtable, colsums)
        }
        
        ## Add Frequency Table to List
        freqtable.list[[i]] <- freqtable

        ## Write CSV
        if (output.csv == TRUE){
            
            fwrite(freqtable,
                   paste0(outputdir,
                          prefix,
                          varname,
                          ".csv"),
                   na = "NA")

        }

        ## Output Kable
        if (output.kable == TRUE){

            cat("\n------------------------------------------------\n")
            cat(paste0("Frequency Table for Variable:   ", varname, "\n"))
            cat("------------------------------------------------\n")
            cat(paste0("\n ",
                       x[, .N, keyby=c(paste0(varname))][,.N],
                       " unique value(s) detected.\n\n"))

            
            print(kable(freqtable,
                        format = "latex",
                        align = "r",
                        booktabs=TRUE,
                        longtable=TRUE) %>% kable_styling(latex_options = "repeat_header"))
        }
    }

    ## Return List of Frequency Tables
    if (output.list == TRUE){
        return(freqtable.list)
    }
}




#'# f.show.values: Show example values of data.table as kable
#' This function functions much like the standard data.table print behaviour, only that the result is a kable table. This is is useful when Listings throws an error for unknown symbols, as kable tables are processed with regular LaTeX, not listings.


#'@param x A data.table for which example values are to be shown.

f.show.values <- function(x){
    length <- length(x)
    rows.final <- seq((x[,.N] - 4), x[,.N], 1)

    for (index.begin in seq(1, length, 4)){
        index.end <- index.begin + 3

        if(index.end > length){
            index.end <- length
        }

        print(kable(x[c(1:5,
                        rows.final),
                      index.begin:index.end],
              format = "latex",
              align = c("p{4cm}"),
              booktabs=TRUE,
              longtable=TRUE))
    }
}






#'# f.dopar.pagenums: Parallelized Computation of the length (in pages) of PDF files
#' This function computes the maximum number of pages for each PDF file. Ideally used with sum() to get the total number of pages of all PDF files in a folder. 


#+
#'## Required Arguments


#' @param x A vector of PDF filenames. Should be located in the working directory.


#+
#'## Required: Cluster
#' WARNING: You must have registered a cluster in advance for this function to work. For example like this:

# cl <- makeForkCluster(detectCores())
# registerDoParallel(cl)


#'## Required Packages

library(doParallel)
library(pdftools)

#'## Function

f.dopar.pagenums <- function(x){

    pagenums <- foreach(filename = x,
                        .combine = 'c',
                        .errorhandling = 'remove',
                        .inorder = FALSE) %dopar% {
                            pdf_length(filename)
                        }
    return(pagenums)

}








#'# f.dopar.pdfextract: Parallelized Extraction of text from PDF files
#' This function parallelizes the extraction of text from each PDF file and saves the results as TXT files. Only the file extension is modified.


#+
#'## Required Arguments


#' @param x A vector of PDF filenames. Should be located in the working directory.


#+
#'## Required: Cluster
#' WARNING: You must have registered a cluster in advance for this function to work. For example like this:

# cl <- makeForkCluster(detectCores())
# registerDoParallel(cl)


#'## Required Packages

library(doParallel)
library(pdftools)

#'## Function

f.dopar.pdfextract <- function(x){

    newnames <- gsub("\\.pdf",
                     "\\.txt",
                     x)

    out <- foreach(i = seq_along(x),
                   .errorhandling = 'pass') %dopar% {            

                       ## Extract text layer from PDF
                       pdf.extracted <- pdf_text(x[i])

                       ## Write TXT to Disk
                       write.table(pdf.extracted,
                                   newnames[i],
                                   quote = FALSE,
                                   row.names = FALSE,
                                   col.names = FALSE)
                   }
    return(length(out))

}






#'# f.dopar.multihashes
#' This function parallelizes computation of both SHA2-256 and SHA3-512 hashes for an arbitrary number of files. It returns a data frame of file names, SHA2-256 hashes and SHA3-512 hashes.


#+
#'## Required Arguments


#' @param x A vector of filenames. Should be located in the working directory.


#+
#'## Required: OpenSSL System Library
#' The function requires the existence of the OpenSSL library on the system. This is because the openssl package for R does not provide SHA 3 capabilities yet.

#+
#'## Required: Cluster
#' WARNING: You must have registered a cluster in advance for this function to work. For example like this:

# cl <- makeForkCluster(detectCores())
# registerDoParallel(cl)

f.dopar.multihashes <- function(x){

    multihashes <- foreach(filename = x,
                           .errorhandling = 'pass',
                           .combine = 'rbind') %dopar% {
                               
                               sha2.256 <- system2("openssl",
                                                   paste("sha256",
                                                         filename),
                                                   stdout = TRUE)
                               
                               sha2.256 <- gsub("^.*\\= ",
                                                "",
                                                sha2.256)
                               
                               sha3.512 <- system2("openssl",
                                                   paste("sha3-512",
                                                         filename),
                                                   stdout = TRUE)
                               
                               sha3.512 <- gsub("^.*\\= ",
                                                "",
                                                sha3.512)
                               
                               out <- data.frame(filename,
                                                 sha2.256,
                                                 sha3.512)
                               return(out)
                           }
    return(multihashes)
    
}






#'# Miscellaneous Functions



#+
#'## f.randomID: Add random ID to Data Tables
#' Adds a randomID for each row to a Data Table object via in-place modification.


#' @param x A data.table.

f.randomID <- function(x){

    x[, randomID := {
        randomID <- sample(.N, .N)
        list(randomID)
        }]
    }





#+
#'## f.linkextract: Extract Links from HTML
#' This function extracts all links (i.e. href attributes of <a> tags) from an arbitrary HTML document. Returns "NA" if there is an error.
#'

#' @param URL A valid URL.

 library(rvest)

f.linkextract <- function(URL){
    tryCatch({
        read_html(URL) %>%
            html_nodes("a")%>%
            html_attr('href')},
        error=function(cond) {
            return(NA)}
        )
}






#'## f.year.iso: Tranform Double-Digit Years to ISO Year Format
#' This function transforms double-digit years (YY) to four-digit years as per the ISO standard (YYYY). It is based on the assumption that years above a certain boundary belong to the 20th century and years at or below that boundary belong to the 21st century.
#'
#'
#'
#' @param inputyear A vector of two-digit years.
#' @param boundary The boundary year. Default is 50 (=1950). 


f.year.iso <- function(inputyear, boundary=50){
    ifelse(inputyear > boundary,
           1900+inputyear,
           2000+inputyear)
}


#'## f.age: Calculate human age
#' Calculates age from date of birth to target date. Vectorized.

#' @param birthdate Date of birth
#' @param target Target date

f.age <- function(birthdate, target){
    if (is.na(birthdate) | is.na(target)){
        return(NA)
    }else{
        age <- year(target) - year(birthdate) - 1
        if (month(target) > month(birthdate)){
            age <- age + 1
        }
        if (month(target) == month(birthdate) && mday(target) >= mday(birthdate)){
            age <- age + 1
        }
        return(age)
    }
}

f.age <- Vectorize(f.age)





#'## f.empty.list.NA: Replaces empty elements in list with "NA"
f.list.empty.NA <- function(x) if (length(x) == 0) NA_character_ else paste(x, collapse = " ")





#'## f.empty.vector.NA: Replaces empty elements in vector with "NA"
f.vec.empty.NA <- function(x) gsub("^$", "NA", x)

